/*************************************************************************
MGLApp.cpp

Clip Earth Demo. Demo for "User Clip Planes" section

Author   :   Andrei Gradinari
Based on Dave Astle's application template

Written for More OpenGL Game Programming
August, 2005
*************************************************************************/

#ifdef WIN32
	#define WIN32_LEAN_AND_MEAN
	#define WIN32_EXTRA_LEAN
	#include <windows.h>
#endif

#include "MGLApp.h"
using namespace MGL;

MGLApp::MGLApp()
{
	m_dElapTime = 0;
}

MGLApp::~MGLApp()
{
}

bool MGLApp::Init()
{	
// EARTH TEXTURE
	TargaImage img;
	img.Load("data/earth.tga");
	img.FlipVertical();
	
	int width = img.GetWidth();
	int height = img.GetHeight();

	glGenTextures(1, &m_iEarthTexture);
	glBindTexture(GL_TEXTURE_2D, m_iEarthTexture);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
	
	gluBuild2DMipmaps(GL_TEXTURE_2D, 3, width, height, GL_RGB, GL_UNSIGNED_BYTE, img.GetImage() );

// EARTH MODEL	
	m_pEarth[0] = CPrimitiveFactory::LoadPrimitive("Data/core.txt");
	m_pEarth[1] = CPrimitiveFactory::LoadPrimitive("Data/middle.txt");
	m_pEarth[2] = CPrimitiveFactory::LoadPrimitive("Data/surface.txt");
	for(int i=0; i<3; i++) if (m_pEarth[i] == NULL) return false;

	return true;
}

bool MGLApp::Shutdown()
{
	return true;
}

void MGLApp::SetupProjection(int width, int height)
{
	// avoid div by zero
	if( height == 0 )
		height = 1;					

	// set up GL viewport
	glViewport( 0, 0, width, height );

	// set up a perspective projection
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 45.0f, (GLfloat)width/(GLfloat)height, 1.0f, 1000.0f );

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE); glCullFace(GL_BACK);
}

void MGLApp::Update( float dt )
{
	m_dElapTime += dt;
}

void setup_earth_material(int obj_index)
{
	vec4 diff, spec, ambt = vec4(.4f, .4f, .4f, 1.0f), emis = vec4(0,0,0,0);
	float shine = 0;
	bool b_textured = false;
    
	switch(obj_index)
	{
		case 0: // core
				diff = vec4(0.5f, 0.3f, 0.15f, 1.0f);
				emis = vec4(1.0f, 0.6f, .3f, 1.0f); 
				spec = diff;
				shine = 25;
				break;	
	
		case 1: // mantle
				diff = vec4(1, .8f, .4f, 1.0f); 
				ambt = diff;		
				spec = diff * 0.5f;
				emis = diff * 0.5f;
				shine = 15;
				break;	

		case 2: // surface
				diff = vec4(0.3f, 0.3f, 0.6f, 1.0f); 
				emis = vec4(0.15f, 0.15f, 0.3f, 1.0f); 
				spec = diff * 0.3f;
				break;

		case 3: // earth surface textured
				diff = vec4(0.7f, 0.7f, 0.7f, 1.0f); 
				spec = vec4(1.0f, 1.0f, 1.0f, 1.0f); 
				b_textured = true;
				shine = 50;
	}

	glMaterialf(GL_FRONT, GL_SHININESS, shine);
	
	glMaterialfv(GL_FRONT, GL_AMBIENT, (GLfloat*)&ambt);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, (GLfloat*)&diff);
	glMaterialfv(GL_FRONT, GL_SPECULAR,(GLfloat*)&spec);
	glMaterialfv(GL_FRONT, GL_EMISSION,(GLfloat*)&emis);

	if(b_textured) glEnable(GL_TEXTURE_2D);
	else glDisable(GL_TEXTURE_2D);
}

void MGLApp::Render()
{
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	// setup camera
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	gluLookAt( 10.0f, 10.0f, 15.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f);
	GLfloat view_mat[16];
	glGetFloatv(GL_MODELVIEW_MATRIX, view_mat);
	
	// setup lighting
	GLfloat pos[4] = {-0.5f, 0.4f, 1.0f, 0.0f};
	GLfloat diff[4] = {1.0f, 1.0f, 1.0f, 1.0f};
	GLfloat spec[4] = {1.0f, 1.0f, 1.0f, 1.0f};

	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHTING);
	glLightfv(GL_LIGHT0, GL_POSITION, pos);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, diff);
	glLightfv(GL_LIGHT0, GL_SPECULAR, spec);
			
	// setup clip planes
	Vector3 verts[3][4] = { // 3 sets of 4 points. Each set defines a plane.
		Vector3(0,0,1), Vector3(0,1,0), Vector3(0,0,-1), Vector3(0,-1,0), 
		Vector3(1,0,0), Vector3(0,0,1), Vector3(-1,0,0), Vector3(0,0,-1),
		Vector3(1,0,0), Vector3(0,-1,0), Vector3(-1,0,0), Vector3(0,1,0)
	};

	GLdouble eq[3][4]; // 3 plane equations
	
	for(int k=0; k<3; k++) get_plane_equation(verts[k][0], verts[k][1], verts[k][2], eq[k]);
	
	// imortant to define plane equations here before glRotate.
	// If you define them after glRotate, clip planes will rotate together with object 
	for( k=0; k<3; k++) glClipPlane(GL_CLIP_PLANE0+k, eq[k]);

	glRotatef( 20.f * (float)m_dElapTime, 0.f, 1.f, 0.f );
	glRotatef( -90, 0.1f, 0.0f, 0.0f);
	
	for( k=0; k<3; k++)
		for(int i=1; i<3; i++)
		{
			if(k==1 && i==2) continue; // comment this string and see the difference
			
			glEnable(GL_CLIP_PLANE0+k);
			glClear(GL_STENCIL_BUFFER_BIT);
			
			// multipass rendering
			glDisable(GL_DEPTH_TEST);
			glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
			glEnable(GL_STENCIL_TEST);

			// first pass: back face increment
			glStencilFunc(GL_ALWAYS, 0, 0);
			glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);
			glCullFace(GL_FRONT);
			m_pEarth[i]->Draw( false, false, false, false );

			// second pass: front face decrement
			glStencilOp(GL_KEEP, GL_KEEP, GL_DECR);
			glCullFace(GL_BACK);
			m_pEarth[i]->Draw( false, false, false, false );

			// drawing clip planes masked by stencil buffer content
			glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
			glEnable(GL_DEPTH_TEST);
			glDisable(GL_CLIP_PLANE0+k);
			
			setup_earth_material(i);
			
			glStencilFunc(GL_NOTEQUAL, 0, ~0);

			// rendering clip edges to frame buffer
			glPushMatrix();
			glLoadMatrixf(view_mat);	// Clip plane had been set up after view_mat space, thus we need to lo load
										// view_mat matrix before rendering clip plane
				glBegin(GL_QUADS);
				Vector3 nrml;
				nrml = Vector3(verts[k][2], verts[k][1]).CrossProduct(Vector3(verts[k][1], verts[k][0])); 
				nrml.Normalize();
				glNormal3f(nrml.x, nrml.y, nrml.z); //We want plane to be lit properly, thus we need to specify it's normal.
				for(int j=3; j>=0; j--)
					// We need plane cover clip edge area thus multiply by 10 all coords 
					glVertex3f(verts[k][j].x*10, verts[k][j].y*10, verts[k][j].z*10);
				glEnd();
			glPopMatrix();
			
			// rendering geometry to frame buffer			
			glDisable(GL_STENCIL_TEST);
			glEnable(GL_CLIP_PLANE0+k);
			setup_earth_material( i==2 ? 3 : i);
			m_pEarth[i]->Draw( false, true, false, true );
			glDisable(GL_CLIP_PLANE0+k);
		}
	
	// drawing Earth's core
	setup_earth_material(0);
	m_pEarth[0]->Draw( false, true, false, false );
}


