/*************************************************************************
MGLSDL.h

Setup code for a basic SDL application.

Author   :   Dave Astle

Written for More OpenGL Game Programming
*************************************************************************/

#include "sdl.h"
#include "MGLApp.h"
#include <cstdlib>

bool g_useVertexArrays = false;
bool g_useBuffers = false;

const float FPS_UPDATE_FREQUENCY = 2.0f;

void SetGLAttributes()
{
  SDL_GL_SetAttribute(SDL_GL_BUFFER_SIZE, MGL_COLOR_BITS);
  SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, MGL_DEPTH_BITS);
  SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, MGL_STENCIL_BITS);
  SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
//  SDL_GL_SetAttribute(SDL_GL_MULTISAMPLEBUFFERS, 1);
//  SDL_GL_SetAttribute(SDL_GL_MULTISAMPLESAMPLES, 4);
}


// returns false if a quit event is received
bool HandleEvent(MGLApp* pApp, const SDL_Event& event)
{
  switch (event.type)
  {
  case SDL_QUIT:
    return false;
  case SDL_KEYDOWN:
    {
      switch (event.key.keysym.sym)
      {
      case SDLK_ESCAPE:
        return false;
      case SDLK_LEFT:
        pApp->RotateLeft();
        break;
      case SDLK_RIGHT:
        pApp->RotateRight();
        break;
      }
    } break;

  case SDL_KEYUP:
    {
      switch (event.key.keysym.sym)
      {
      case SDLK_SPACE:
        g_useVertexArrays = pApp->ToggleVertexArrays();
        break;
      case SDLK_TAB:
        g_useBuffers = pApp->ToggleBuffers();
        break;
      }
    } break;

  case SDL_VIDEORESIZE:
    pApp->SetupProjection(event.resize.w, event.resize.h);
    break;
  }

  return true;
}


void DisplayTitle(float fps)
{
  static char message[255];
  sprintf(message, "%s | FPS: %.0f Vertex Arrays: %s Vertex Buffer Objects: %s ",
    MGL_APP_NAME, fps, g_useVertexArrays ? "on" : "off", g_useBuffers ? "on" : "off");
  SDL_WM_SetCaption(message, message);
}

int main(int argc, char* argv[])
{
  SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER);
  atexit(SDL_Quit);

  DisplayTitle(0.0f);

  SDL_EnableKeyRepeat(1, SDL_DEFAULT_REPEAT_INTERVAL);
  SetGLAttributes();

  Uint32 displayFlags = SDL_OPENGL;

  if (MGL_FORCE_FULLSCREEN)
  {
    displayFlags |= SDL_FULLSCREEN;
  }
  else
  {
    displayFlags |= SDL_RESIZABLE;
  }

  SDL_Surface* pDisplay = SDL_SetVideoMode(MGL_WINDOW_WIDTH, MGL_WINDOW_HEIGHT, MGL_COLOR_BITS, displayFlags);

  if (pDisplay == NULL)
  {
    fprintf(stderr, "Couldn't set GL mode: %s\n", SDL_GetError());
    exit(0);
  }

  MGLApp* pApp = new MGLApp();
  pApp->Init();
  pApp->SetupProjection(MGL_WINDOW_WIDTH, MGL_WINDOW_HEIGHT);

  Uint32 lastUpdate, lastTime, frame = 0;
  lastUpdate = lastTime = SDL_GetTicks();

  for (;;)
  {
    SDL_Event event;
    Uint32 currTime;
    

    if (SDL_PollEvent(&event) == 0)
    {
      currTime = SDL_GetTicks();
      pApp->Update((currTime - lastTime)/1000.0f);
      lastTime = currTime;
      pApp->Render();

      frame++;
      if ((currTime - lastUpdate)/1000.0f > FPS_UPDATE_FREQUENCY)
      {
        DisplayTitle(frame/FPS_UPDATE_FREQUENCY);
        lastUpdate = currTime;
        frame = 0;
      }
      SDL_GL_SwapBuffers();
    }
    else if (!HandleEvent(pApp, event))
      break;
  }
  pApp->Shutdown();
  delete pApp;

  return 0;
}